#include <winsock2.h>
#include <iphlpapi.h>
#include <stdio.h>
#include <stdlib.h>

#include "MachineId.h"
#include "Logger.h"

// Link with Iphlpapi.lib
#pragma comment(lib, "IPHLPAPI.lib")

#define WORKING_BUFFER_SIZE 15000
#define MAX_TRIES 3

#define MALLOC(x) HeapAlloc(GetProcessHeap(), 0, (x))
#define FREE(x) HeapFree(GetProcessHeap(), 0, (x))

static char* getEnv(const char* envName);

static const char* ENV_NAME = "PARASOFT_SUPPORT_NET_INTERFACES";
static const char* ENV_DEBUG = getEnv("PARASOFT_DEBUG_NET_INTERFACES");
static const char* ENV_DEBUG_FILE = getEnv("PARASOFT_DEBUG_NET_INTERFACES_FILE");
static PIP_ADAPTER_ADDRESSES SELECTED_ADAPTER = NULL;
static const bool IS_DEBUG_FILE = ENV_DEBUG_FILE != NULL && strcmp(ENV_DEBUG_FILE, "") != 0;
static const bool DEBUG = IS_DEBUG_FILE || (ENV_DEBUG != NULL && strcmp(ENV_DEBUG, "true") == 0);

PIP_ADAPTER_ADDRESSES findAdapter(PIP_ADAPTER_ADDRESSES pAddresses, DWORD type);
PIP_ADAPTER_ADDRESSES findAdapter(PIP_ADAPTER_ADDRESSES pAddresses, DWORD type, char * envAdress);
int generateMachineId(unsigned char PhysicalAddress[], int PhysicalAddressLength);
static void printDebug(const char* format, ...);
bool addressExist(char * envAddress, PIP_ADAPTER_ADDRESSES pAddress);
int convertToDecimal(char x, char y);
int charToInt(char x);
PIP_ADAPTER_ADDRESSES scanAdapters(PIP_ADAPTER_ADDRESSES pBestAddresses, PIP_ADAPTER_ADDRESSES pAddresses);
void printInfoAboutSelected(PIP_ADAPTER_ADDRESSES pAddress);
void printInfoAboutAdapter(PIP_ADAPTER_ADDRESSES pAddress);
void printInfoAboutCurrent(PIP_ADAPTER_ADDRESSES pAddress);

MachineId::MachineId(void)
{
}

MachineId::~MachineId(void)
{
}

// See: http://msdn.microsoft.com/en-us/library/windows/desktop/aa365915%28v=vs.85%29.aspx
int MachineId::getMachineId()
{

	char* envInterface = getEnv(ENV_NAME);

    DWORD dwRetVal = 0;

    // Set the flags to pass to GetAdaptersAddresses
    ULONG flags = GAA_FLAG_INCLUDE_PREFIX;

    // Unspecified address family (both). Other values: AF_INET, AF_INET6
    ULONG family = AF_UNSPEC;

    PIP_ADAPTER_ADDRESSES pAddresses = NULL;
    ULONG outBufLen = 0;
    ULONG Iterations = 0;

    PIP_ADAPTER_ADDRESSES pBestAddresses = NULL;
    // Allocate a 15 KB buffer to start with.
    outBufLen = WORKING_BUFFER_SIZE;
    
    // Gathering adapters from machine using windows api
	do {
        pAddresses = (IP_ADAPTER_ADDRESSES *) MALLOC(outBufLen);
        if (pAddresses == NULL) {
            setErrorMessage("Memory allocation failed for IP_ADAPTER_ADDRESSES struct");
            Logger::logln("Memory allocation failed for IP_ADAPTER_ADDRESSES struct");
            return 0x0;
        }

        // The order in which adapters appear in the list returned by this function
        // can be controlled from the Network Connections folder: select the Advanced Settings menu item from the Advanced menu
        printDebug("[DEBUG] Getting adapters addresses, size for Addresses: %d \n", outBufLen);
		dwRetVal = GetAdaptersAddresses(family, flags, NULL, pAddresses, &outBufLen);
        if (dwRetVal == ERROR_BUFFER_OVERFLOW) {
			printDebug("[DEBUG]ERROR BUFFER OVERFLOW catched, size for Addresses: %d \n", outBufLen);
            FREE(pAddresses);
            pAddresses = NULL;
        } else {
            break;
        }
        Iterations++;
    } while ((dwRetVal == ERROR_BUFFER_OVERFLOW) && (Iterations < MAX_TRIES));
    
    char error[30];
	sprintf_s(error, sizeof(error),"%d", dwRetVal);
	if (dwRetVal != ERROR_SUCCESS) {
		printDebug("[DEBUG] ERROR DETECTED: %s. Ending machine id calculation \n", error);
		char messageBuffer [100];
        setErrorMessage(messageBuffer);
        Logger::logln(messageBuffer);
		return 0x0;
	}
	
    int machineId = 0;
		
	if (envInterface != 0) {
		// There is environment variable which determines preferred mac address
		printDebug("[DEBUG] Checking physical address from environment variable \n");
		pBestAddresses = findAdapter(pAddresses, IF_TYPE_ETHERNET_CSMACD, envInterface);
		bool baddressExists = SELECTED_ADAPTER != NULL;
		
		if (baddressExists) {
			machineId = generateMachineId(SELECTED_ADAPTER->PhysicalAddress, (int)SELECTED_ADAPTER->PhysicalAddressLength);
			
		} else { 
			if (pBestAddresses) {
				machineId = generateMachineId(pBestAddresses->PhysicalAddress, (int)pBestAddresses->PhysicalAddressLength);
			}
		}
		
		if (machineId != 0) { 
			printDebug("[SELECTED] Physical Address: ");
			printDebug(envInterface);
			printDebug(" (user-provided) \n");
		}
		
	} else { 
		// There is no environment variable, determining preferred mac address so we are doing standard scanning
		pBestAddresses = scanAdapters(pBestAddresses, pAddresses);
	    if (pBestAddresses) {
	        machineId = generateMachineId(pBestAddresses->PhysicalAddress, (int) pBestAddresses->PhysicalAddressLength);
	    } else {
	        Logger::logln("Cannot find network adapter. Run ipconfig /all to list known adapters.");
	    }
	}

    if (pAddresses) {
        FREE(pAddresses);
    }

    return machineId;
}

PIP_ADAPTER_ADDRESSES findAdapter(PIP_ADAPTER_ADDRESSES pAddresses, DWORD type)
{
	return findAdapter(pAddresses, type, NULL);
}

PIP_ADAPTER_ADDRESSES findAdapter(PIP_ADAPTER_ADDRESSES pAddresses, DWORD type, char * envAdress)
{
    PWCHAR pName = NULL;
    PIP_ADAPTER_ADDRESSES pAddress = NULL;
    PIP_ADAPTER_ADDRESSES pCurrAddresses = pAddresses;
	boolean bcheckEnvAddress = envAdress != NULL;
	int index = 0;
	const int numberOfBytesInMACAddress = 6;
	if (bcheckEnvAddress) {
		printDebug("[DEBUG] Checking if address: %s from environment variable exists \n", envAdress);
	}
    while (pCurrAddresses) {
		int pAddressLength = pCurrAddresses->PhysicalAddressLength;
		if (pAddressLength != numberOfBytesInMACAddress) {
			printDebug("[WARN] Received incorrect physical address, so skipping it. Received Length: %d \n", pAddressLength);
			pCurrAddresses = pCurrAddresses->Next;
			continue;
		}
    	unsigned char * MAC = pCurrAddresses->PhysicalAddress;
    	if (!MAC) { 
    		printDebug("[WARN] Received incorrect physical address");
    		pCurrAddresses = pCurrAddresses->Next;
    		continue;
    	} 
    	printInfoAboutCurrent(pCurrAddresses);
		if (bcheckEnvAddress) {
			for (int i = 0; i < pAddressLength; i++) {
				unsigned char octet = MAC[i];
				char x = envAdress[index];
				index++;
				char y = envAdress[index];
				index++;
				if (envAdress[index] == ':' || envAdress[index] == '-') {
					index++;
				}
				if (octet != convertToDecimal(x, y)) {
					index = 0;
					break;
				}
			}
			if (index != 0) {
				SELECTED_ADAPTER = pCurrAddresses;
				printDebug("[DEBUG] Adapter %s exists, accepting \n", envAdress);
				printInfoAboutSelected(SELECTED_ADAPTER);
				return SELECTED_ADAPTER;
			}
		}
        if ((type == -1) || (pCurrAddresses->IfType == type)) {
            // Sort adapters by name (pr108045)
            if ((pCurrAddresses->PhysicalAddressLength != 0) && (pCurrAddresses->FriendlyName != NULL)) {
                if ((pName == NULL) || wcscmp(pCurrAddresses->FriendlyName, pName) < 0) {
                    pName = pCurrAddresses->FriendlyName;
                    pAddress = pCurrAddresses;
                }
            }
        }
        pCurrAddresses = pCurrAddresses->Next;
    }
    if (bcheckEnvAddress) { 
    	printDebug("[WARN] The network card with physical address set in PARASOFT_SUPPORT_NET_INTERFACE variable has not been found. Ensure the variable is properly set \n");
		printDebug("[DEBUG] Using adapter received during standard scanning \n");
    }
    if (!pAddress) {
		printDebug("No adapters available");
		return 0x0;
	}
	printInfoAboutSelected(pAddress);
    return pAddress;
}

int generateMachineId(unsigned char PhysicalAddress[], int PhysicalAddressLength)
{
    // Copied from Linux machine id calculation
    union U {
        unsigned char c[4];
        unsigned int  n;
    } u;

    unsigned int cSum = 0;
    for (int i = 0; i < PhysicalAddressLength; i++) {
        cSum += PhysicalAddress[i];
    }
    if (PhysicalAddressLength > 1)
        u.c[0] = PhysicalAddress[1];
    if (PhysicalAddressLength > 3)
        u.c[1] = PhysicalAddress[3];
    if (PhysicalAddressLength > 5)
        u.c[2] = PhysicalAddress[5];
    u.c[3] = (unsigned char)cSum;
    return u.n;
}

static char* getEnv(const char* envName) {
    char * envVariable;
    _dupenv_s(&envVariable, 0, envName);
    return envVariable;
}

static void printDebug(const char* format, ...)
{
	if (!DEBUG) {
		return;
	}

    va_list argptr;
    va_start(argptr, format);

	if (IS_DEBUG_FILE) {
		FILE * debugFile;
        fopen_s(&debugFile, ENV_DEBUG_FILE, "a");
		if (debugFile != NULL) {
		    vfprintf(debugFile, format, argptr);
			fclose(debugFile);
		} else {
		    printf("[Wrong PARASOFT_DEBUG_NET_INTERFACES_FILE] %s\n", ENV_DEBUG_FILE);
		    vfprintf(stdout, format, argptr);
		}
	} else {
	    vfprintf(stdout, format, argptr);
	}

    va_end(argptr);

}

int convertToDecimal(char x, char y)
{
	int count1 = charToInt(x);
	int count2 = charToInt(y);

	return (count1 * 16) + count2;
}

int charToInt(char x)
{
	if (isdigit(x)) {
		return x - '0';
	}
	char hexadecimalChars[16] = { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F' };
	for (int i = 0; i < sizeof(hexadecimalChars); i++) {
		if (x == hexadecimalChars[i]) {
			return i;
		}
	}
	return x;
}

PIP_ADAPTER_ADDRESSES scanAdapters(PIP_ADAPTER_ADDRESSES pBestAddresses, PIP_ADAPTER_ADDRESSES pAddresses)
{
	// Ethernet card first
        printDebug("[DEBUG] Scanning ethernet adapters \n");
        pBestAddresses = findAdapter(pAddresses, IF_TYPE_ETHERNET_CSMACD);
        
        // Wireless card next
        if (!pBestAddresses) {
        	printDebug("[DEBUG] Scanning wireless adapters \n");
            pBestAddresses = findAdapter(pAddresses, IF_TYPE_IEEE80211);
        }
        
        // Other network card
        if (!pBestAddresses) {
        	printDebug("[Scanning] other adapters \n");
            pBestAddresses = findAdapter(pAddresses, (DWORD)-1);
        }
        return pBestAddresses;
        
}

void printInfoAboutSelected(PIP_ADAPTER_ADDRESSES pAddress)
{
	printDebug("[SELECTED] ");
	printInfoAboutAdapter(pAddress);
}

void printInfoAboutCurrent(PIP_ADAPTER_ADDRESSES pAddress)
{
	printDebug("[DEBUG] ");
	printInfoAboutAdapter(pAddress);
}

void printInfoAboutAdapter(PIP_ADAPTER_ADDRESSES pAddress)
{

	unsigned char * SELECTED = pAddress->PhysicalAddress;
	printDebug("%x-%x-%x-%x-%x-%x", SELECTED[0], SELECTED[1], SELECTED[2], SELECTED[3], SELECTED[4], SELECTED[5]);
	printDebug("\t");
	printDebug("Interface type: %d", pAddress->IfType);
	printDebug("\n");
}
